// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Xml;
using System.IO;

using IPattern = PiDuce.Common.ISchema;

namespace PiDuce.Common
{
	public interface IProcessVisitor
	{
		void VisitNilProcess(NilProcess p);
		void VisitSpawnProcess(SpawnProcess p);
		void VisitOutputProcess(OutputProcess p);
		void VisitInputProcess(InputProcess p);
		void VisitJoinProcess(JoinProcess p);
		void VisitMatchProcess(MatchProcess p);
		void VisitImportProcess(ImportProcess p);
		void VisitNewProcess(NewProcess p);
		void VisitSchemaProcess(SchemaProcess p);
	}

	public interface IProcess
	{
		void Accept(IProcessVisitor visitor);
	}

	public class NilProcess : IProcess
	{
		public NilProcess()
		{ }

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitNilProcess(this); }
	}

	public class OutputProcess : IProcess
	{
		private readonly IExpression destination;
		private readonly IExpression expression;

		public OutputProcess(IExpression destination, IExpression expression)
		{
			this.destination = destination;
			this.expression = expression;
		}

		public IExpression Destination { get { return destination; } }
		public IExpression Expression { get { return expression; } }

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitOutputProcess(this); }
	}

	public class InputProcess : IProcess
	{
    private readonly IExpression source;
		private readonly IPattern pattern;
		private readonly IProcess continuation;
		private readonly bool star;  //true => x?*(F).P  | false => x?(F).P

		public IExpression Source { get { return source; } }
		public IPattern Pattern { get { return pattern; } }
		public IProcess Continuation { get { return continuation; } }
		public bool Service { get { return star; } }
						
		public InputProcess(IExpression source, bool star, IPattern pattern, IProcess continuation)
		{
			this.source = source;
			this.pattern = pattern;
			this.continuation = continuation;
			this.star = star;
		}

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitInputProcess(this); }
	}

	public class SpawnProcess : IProcess
	{
		private readonly IProcess left;
		private readonly IProcess right;
		private int parentEnvSize;
		private int leftEnvSize;

		public IProcess Left { get { return left; } }
		public IProcess Right { get { return right; } }
		public int ParentEnvSize { get { return parentEnvSize; } }
		public int LocalEnvSize { get { return leftEnvSize - parentEnvSize; } }

		public SpawnProcess(IProcess left, IProcess right)
		{
			this.left = left;
			this.right = right;
			this.parentEnvSize = 0;
			this.leftEnvSize = 0;
		}

		public void SetEnvSize(int parentEnvSize, int leftEnvSize)
		{
			Debug.Assert(leftEnvSize >= parentEnvSize);
			this.parentEnvSize = parentEnvSize;
			this.leftEnvSize = leftEnvSize;
		}

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitSpawnProcess(this); }
	}

	public class MatchProcess : IProcess
	{
		public class Branch
		{
			private readonly IPattern pattern;
			private readonly IExpression condition;
			private readonly IProcess process;

			public IPattern Pattern { get { return pattern; } }
			public IExpression Condition { get { return condition; } }
			public IProcess Process { get { return process; } }

			public Branch(IPattern pattern, IExpression condition, IProcess process)
			{
				this.pattern = pattern;
				this.condition = condition;
				this.process = process;
			}
		}

		private readonly IExpression expression;
		private readonly IList<Branch> branches;

		public IExpression Expression { get { return expression; } }
		public IList<Branch> Branches { get { return branches; } }

		public MatchProcess(IExpression expression, IList<Branch> branches)
		{
			this.expression = expression;
			this.branches = branches;
		}

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitMatchProcess(this); }
	}

	public class ImportProcess : IProcess
	{
    private readonly string name;
    private readonly IDictionary<string, ISchema> operations;
    private readonly IProcess continuation;
    private readonly string wsdl;
    private IEntry entry;

    public ImportProcess(string name, IDictionary<string, ISchema> operations, string wsdl, IProcess continuation)
		{
			this.name = name;
      this.operations = operations;
      this.entry = null;
			this.wsdl = wsdl;
			this.continuation = continuation;
		}

		public string Name { get { return name; } }
    public IDictionary<string, ISchema> Operations { get { return operations; } }
    public IEntry Entry {
			get { return entry; }
			set { entry = value; }
		}
		public IProcess Continuation { get { return continuation; } }
		public string Wsdl { get { return wsdl; } }

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitImportProcess(this); }
	}

	public class NewProcess : IProcess
	{
		private readonly string name;
    private readonly IDictionary<string, ISchema> operations;
    private readonly IDictionary<string, ChannelType.CAPABILITY> exportedCapabilities;
		private readonly IProcess continuation;
		private readonly string location;
    private IEntry entry;

		public string Name { get { return name; } }
    public IDictionary<string, ISchema> Operations { get { return operations; } }
    public IDictionary<string, ChannelType.CAPABILITY> ExportedCapabilites { get { return exportedCapabilities; } }
		public IProcess Continuation { get { return continuation; } }
		public string ChannelLocation { get { return location; } }
    public IEntry Entry {
			get { return entry; }
			set { entry = value; }
		}

    private static ISchema ReplaceCapability(ISchema s, out ChannelType.CAPABILITY oldCapability)
    {
      if (s is ChannelSchema) {
        ChannelSchema t = (ChannelSchema)s;
        oldCapability = t.Capability;
        return new ChannelSchema(t.Content, ChannelType.CAPABILITY.INOUT);
      } else if (s is FunctionSchema) {
        FunctionSchema t = (FunctionSchema)s;
        oldCapability = t.Capability;
        return new FunctionSchema(t.Input, t.Output, ChannelType.CAPABILITY.INOUT);
      } else if (s is ConstantSchema) {
        ConstantSchema t = (ConstantSchema)s;
        // by the time we get here names must have been resolved
        Debug.Assert(t.Entry != null);
        return ReplaceCapability(t.Entry.Schema, out oldCapability);
      } else {
        Debug.Assert(false);
        oldCapability = ChannelType.CAPABILITY.INOUT; // not to make the compiler complain
        return null;
      }
    }

    public NewProcess(string name, IDictionary<string, ISchema> operations, string location, IProcess continuation)
		{
			this.name = name;
      this.operations = new Dictionary<string, ISchema>();
      this.exportedCapabilities = new Dictionary<string, ChannelType.CAPABILITY>();
      foreach (KeyValuePair<string, ISchema> operation in operations) {
        ChannelType.CAPABILITY oldCapability;
        this.operations.Add(operation.Key, ReplaceCapability(operation.Value, out oldCapability));
        this.exportedCapabilities.Add(operation.Key, oldCapability);
      }
      this.entry = null;
			this.location = location;
			this.continuation = continuation;
		}

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitNewProcess(this); }
	}

	public class SchemaProcess : IProcess
	{
		public class Declaration
		{
			private readonly string name;
			private readonly ISchema schema;
			private IEntry entry;

			public string Name { get { return name; } }
			public ISchema Schema { get { return schema; } }
			public IEntry Entry {
				get { return entry; }
				set { entry = value; }
			}

			public Declaration(string name, ISchema schema)
			{
				this.name = name;
				this.schema = schema;
			}
		}

		private IList<Declaration> decls;
		private readonly IProcess continuation;

		public IList<Declaration> Declarations { get { return decls; } }
		public IProcess Continuation { get { return continuation; } }

		public SchemaProcess(IList<Declaration> decls, IProcess continuation)
		{
			this.decls = decls;
			this.continuation = continuation;
		}

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitSchemaProcess(this); }
	}

	// BEGIN: JOIN PATTERNS
  // an input channel with associated pattern occurring in an
  // input definition
  public class JoinPatternAtom
  {
		private readonly IExpression source;
    private readonly IPattern pattern;
		private int index;
    private IEntry entry;

		public IExpression Source { get { return source; } }
		public IPattern Pattern { get { return pattern; } }
		public int Index { 
			get { return index; }
			set { index = value; }
		}
    public IEntry Entry
    {
      get { return entry; }
      set { entry = value; }
    }

		public JoinPatternAtom(IExpression source, IPattern pattern)
			: this(source, pattern, -1)
		{ }

    public JoinPatternAtom(IExpression source, IPattern pattern, int index)
    {
      this.source = source;
      this.pattern = pattern;
			this.index = index;
      this.entry = null;
    }
  }

  // a set of input channels associated with a continuation
  public class JoinPatternRow
  {
    private readonly IList<JoinPatternAtom> atoms;
    private readonly IProcess continuation;

		public IList<JoinPatternAtom> Atoms { get { return atoms; } }
		public IProcess Continuation { get { return continuation; } }

    public JoinPatternRow(IList<JoinPatternAtom> atoms, IProcess continuation)
    {
      this.atoms = atoms;
      this.continuation = continuation;
    }
  }

	public class JoinProcess : IProcess
	{
    private readonly IList<JoinPatternRow> rows;
		private readonly bool service;

		public IList<JoinPatternRow> Rows { get { return rows; } }
		public bool Service { get { return service; } }

		public JoinProcess(bool service, IList<JoinPatternRow> rows)
		{
			this.service = service;
			this.rows = rows;

#if false
			// collect channel names
			IList<string> channels = new List<string>();
			foreach (JoinPatternRow row in rows)
				foreach (JoinPatternAtom atom in row.Atoms)
					if (!channels.Contains(atom.Channel))
						channels.Add(atom.Channel);

			// assign each channel its index
			foreach (JoinPatternRow row in rows)
				foreach (JoinPatternAtom atom in row.Atoms)
					atom.Index = channels.IndexOf(atom.Channel);
#endif
		}

		public void Accept(IProcessVisitor visitor)
		{ visitor.VisitJoinProcess(this); }

  }
}
