// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System.Collections.Generic;
using System.Diagnostics;

namespace PiDuce.Common
{
	public interface IEntry
	{
		string Name { get; }
		ISchema Schema { get; set; }
	}

	public interface IResolver
	{
		IEntry GetEntry(string name);
	}

	public abstract class AbstractEntry : IEntry
	{
		private readonly string name;
		private ISchema schema;

		public AbstractEntry(string name, ISchema schema)
		{
			this.name = name;
			this.schema = schema;
		}

		public string Name { get { return this.name; } }

		public ISchema Schema {
			get { return this.schema; }
			set { this.schema = value; }
		}
	}

	public class SymbolTable<T> : IResolver where T : IEntry
	{
		private readonly IDictionary<string, Bucket> table;
		private readonly Stack<Environment> envStack;

		protected class Bucket
		{
			public readonly string name;
			public Binding current;
	
			public Bucket(string name)
			{
				this.name = name;
				this.current = null;
			}
		}

		protected class Binding
		{
			public readonly Bucket bucket;
			public readonly T entry;
			public readonly Binding previous;

			public Binding(Bucket bucket, T entry, Binding previous)
			{
				this.bucket = bucket;
				this.entry = entry;
				this.previous = previous;
			}
		}

		protected class Environment
		{
			public readonly IList<Binding> bindings = new List<Binding>();

			public bool Has(string name)
			{
				foreach (Binding binding in bindings)
					if (binding.bucket.name == name)
						return true;
				return false;
			}
		}

		public SymbolTable()
		{
			this.table = new Dictionary<string, Bucket>();
			this.envStack = new Stack<Environment>();
			this.envStack.Push(new Environment());
		}

		protected Bucket GetBucket(string name)
		{
			if (table.ContainsKey(name))
				return table[name];
			else {
				Bucket bucket = new Bucket(name);
				table.Add(name, bucket);
				return bucket;
			}
		}

		protected Environment Top {
			get {
				Debug.Assert(envStack.Count > 0);
				return envStack.Peek();
			}
		}

		public void Add(T entry)
		{
			Environment env = Top;
			Bucket bucket = GetBucket(entry.Name);
			Binding binding = new Binding(bucket, entry, bucket.current);
			env.bindings.Add(binding);
			bucket.current = binding;
		}

		public virtual void Enter()
		{
			this.envStack.Push(new Environment());
		}

		public virtual IList<T> Exit()
		{
			List<T> entries = new List<T>();
			foreach (Binding binding in Top.bindings) {
				entries.Add(binding.entry);
				binding.bucket.current = binding.previous;
			}
			this.envStack.Pop();
			entries.Reverse();
			return entries;
		}

		public virtual T Lookup(string name)
		{
			Bucket bucket = GetBucket(name);
			if (bucket.current != null)
				return bucket.current.entry;
			else
				return default(T);
		}

		public IEntry GetEntry(string name)
		{ return Lookup(name); }

		public virtual bool Clash(string name)
		{ return Top.Has(name); }
	}
}
