// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.IO;
using System.Net;
using System.Xml;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using PiDuce.Common;
using PiDuce.Types;

namespace PiDuce.Compiler
{
  /// <summary>
  /// Compiles PiDuce programs producing the XML bytecode
  /// </summary>
  class Compiler
  {
    private static string usage = "Usage: PiDuceCompiler [-w] <filename>"
      + System.Environment.NewLine
      + "Compile a PiDuce program"
      + System.Environment.NewLine + "Options" + System.Environment.NewLine
      + "-w, /w" + "\t verify every import downloading the wsdl interface";
    private static bool checkWsdl = false;

    public static bool CheckWsdl { get { return checkWsdl; } }
    public static string Usage { get { return usage; } }

    [STAThread]
    static void Main(string[] args)
    {			
      String host = String.Empty;
			String filename;
      if (args.Length < 1) {
	  		Console.WriteLine(Compiler.usage);
	  		return;
			}
      if (args.Length == 2)
      {
        if (args[0] == "-w" || args[0] == "/w") checkWsdl = true;
        filename = args[1];
      }
      else filename = args[0];
      if (filename.IndexOf("@") != -1) {
	  		host = filename.Substring(args[0].IndexOf("@") + 1);
	  		filename = filename.Substring(0, args[0].IndexOf("@"));
			}

			CompilerOutput.FileName = filename;
      String input = String.Empty;
      try {
	  		using (StreamReader file = new StreamReader(filename)) {
	      	while (file.Peek() != -1)
		  			input += file.ReadLine() + "\n";
	      	file.Close();
	    	}
			} catch (Exception) {
	  		Console.WriteLine("Error reading file {0}", filename);
	  		return;
			}
      
      IProcess ast = null;
      PiDuceParser prg = new PiDuceParser();		
			SchemaSymbolTable tst = new SchemaSymbolTable();
      try {
        SchemaEntry anyEntry = tst.New("any", Schema.Any);
        SchemaEntry emptyEntry = tst.New("empty", Schema.Empty);
        tst.Add(anyEntry);
        tst.Add(emptyEntry);
 //Console.WriteLine("BEGIN PARSE");
	  		ast = prg.Parse(input);
 //Console.WriteLine("END PARSE");
				ast.Accept(new ResolveNamesProcessVisitor(tst, new ValueSymbolTable()));
			} catch (SyntaxError) {
				return;
			} catch (Exception e) {
	  		Console.WriteLine(e.Message);
	  		return;
			}

			LocationContext locationContext = new LocationContext();
			ExpressionSchemaContext schemaContext = new ExpressionSchemaContext();
			SchemaChecker schemaChecker = new SchemaChecker();
 //Console.WriteLine("BEFORE TYPE CHECKING");
			new CheckProcessVisitor(locationContext, schemaContext, schemaChecker).Check(ast);
 //Console.WriteLine("AFTER TYPE CHECKING");
      if (CompilerOutput.HasErrors) {
	  		CompilerOutput.PrintResult();
	  		return;
			}

			StringWriter swCode = null;
      XmlTextWriter code = null;
      if (host != "") {                
	  		if (!host.StartsWith("http://")) {
	      	CompilerOutput.Message("Must specify a valid location to upload code");
	      	return;
	    	}

				swCode = new StringWriter();
	  		code = new XmlTextWriter(swCode);
			} else {
				int dot = filename.LastIndexOf('.');
				string object_filename;
				if (dot == -1)
					object_filename = filename + ".pdoc";
				else
					object_filename = filename.Substring(0, dot) + ".pdoc"; 

	  		code = new XmlTextWriter(object_filename, System.Text.Encoding.UTF8);
			}

      code.Formatting = Formatting.Indented;
  		code.WriteStartDocument();
      code.WriteStartElement("PiDuce");
      code.WriteAttributeString("xmlns", WebNamespace.PiDuceOpcodeNamespace);
      code.WriteAttributeString("xmlns", WebNamespace.PiDuceValueNamespacePrefix, null, WebNamespace.PiDuceValueNamespace);

			// DUMPING TYPE ENVIRONMENT
      code.WriteStartElement("schemadecl");
			foreach (SchemaEntry entry in tst.Entries) {
				code.WriteStartElement("schemadec");
				if (entry.Name != entry.FreshName)
					code.WriteAttributeString("original-name", entry.Name);
				code.WriteAttributeString("name", entry.FreshName);
				entry.Schema.Accept(new SchemaToXmlVisitor(code));
				code.WriteEndElement(); // </schemadec>
			}
      code.WriteEndElement();
			// DONE DUMPING TYPE ENVIRONMENT

			AllocateContext allocateContext = new AllocateContext();
			new AllocateNamesProcessVisitor(allocateContext).AllocateNames(ast);
      code.WriteStartElement("process");			
			code.WriteAttributeString("local-env-size", allocateContext.Size.ToString());
			new CompileProcessVisitor(schemaContext, code).Compile(ast);
      code.WriteEndElement(); // </process>
      code.WriteEndElement();	// </PiDuce>
      code.Flush();

      if (host != "") {
	  		try {
          WebClient wc = new WebClient();
	      	CompilerOutput.Message("Sending compiled to " + host);
	      	wc.Headers.Add("User-Agent: PiDuce");
	      	wc.Headers.Add("Content-Type: text/xml; charset=utf-8");
	      	wc.Headers.Add("SOAPAction: http://www.cs.unibo.it/PiDuce/LoadCode");
          string bytecode = swCode.ToString();
          int left = bytecode.IndexOf("<?");
          int right = bytecode.IndexOf("?>") + 1;
          bytecode = bytecode.Remove(left, right - left + 1);
          wc.UploadData(host, "POST", System.Text.Encoding.UTF8.GetBytes(XmlUtil.AddSoap(bytecode)));
	      	CompilerOutput.Message("Success sending compiled to " + host);
	    	} catch (Exception e) {
	      	CompilerOutput.Message("Failed sending compiled to " + host + ". " +e.ToString());
	    	}
			}
    }
  }
}
