// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections.Generic;
using PiDuce.Common;

namespace PiDuce.Compiler
{
	public abstract class CompilerAbstractEntry : AbstractEntry
	{
		private bool used;

		public bool Used {
			get { return used; }
			set { used = value; }
		}

		public CompilerAbstractEntry(string name, ISchema schema)
			: base(name, schema)
		{ used = false; }
	}

	public class SchemaEntry : CompilerAbstractEntry
	{
		private readonly string freshName;

		public string FreshName { get { return freshName; } }

		public SchemaEntry(string name, ISchema schema, string freshName)
			: base(name, schema)
		{ this.freshName = freshName; }
	}

	public class ValueEntry : CompilerAbstractEntry
	{
		private int index;
		private readonly bool local;

		public int Index {
			get { return index; }
			set { index = value; }
		}

		public bool Local { get { return local; } }

		public ValueEntry(string name, ISchema schema, bool local)
			: base(name, schema)
		{
			this.index = -1;
			this.local = local;
		}
	}

  public class ServiceEntry : ValueEntry {
    private IDictionary<string, ValueEntry> operations;

    public IDictionary<string, ValueEntry> Operations { get { return operations; } }

    public ServiceEntry(string name, ISchema schema, IDictionary<string, ValueEntry> operations, bool local)
      : base(name, schema, local)
    { this.operations = operations; }
  }

	public abstract class CompilerSymbolTable<T> : SymbolTable<T>
		where T : CompilerAbstractEntry
	{
		public abstract T New(string name, ISchema schema);

    public T Lookup(string name, int location)
    {
      T entry = base.Lookup(name);
    	if (entry == null) {
    	  CompilerOutput.Error("unbound name - " + name, location);
    		return New(name, new ErrorSchema());
    	} else {
				entry.Used = true;
				return entry;
			}
    }

    public override T Lookup(string name)
    { return Lookup(name, -1); }
	}

  public class SchemaSymbolTable : CompilerSymbolTable<SchemaEntry>
  {
    private int nextIndex = 0;
    private ArraySet<string> names = new ArraySet<string>();
		private IList<SchemaEntry> entries = new List<SchemaEntry>();

		public IList<SchemaEntry> Entries { get { return entries; } }

    public override SchemaEntry New(string name, ISchema schema)
    {
			string freshName = names.Contains(name) ? name + "%" + (nextIndex++).ToString() : name;
      SchemaEntry entry = new SchemaEntry(name, schema, freshName);
      names.Add(name);
			entries.Add(entry);
      return entry;
    }
  }

	public class ValueSymbolTable : CompilerSymbolTable<ValueEntry>
	{
		public override ValueEntry New(string name, ISchema schema)
		{ return New(name, schema, false); }

		public ValueEntry New(string name, ISchema schema, bool local)
		{ return new ValueEntry(name, schema, local); }

    public ServiceEntry NewService(string name, ISchema schema, IDictionary<string, ValueEntry> operations, bool local)
    { return new ServiceEntry(name, schema, operations, local); }
	}
}

