// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using PiDuce.Common;

//using Assumption = PiDuce.Common.Pair<PiDuce.Common.ISchema, PiDuce.Common.ISchema>;

namespace PiDuce.Types
{ 
	public class SubtypingChecker
	{
    private readonly EmptynessChecker ec;
    private readonly DeterminednessChecker dc;
		private readonly HandlesOfSchemaVisitor handlesOf;
		private readonly IDictionary<Assumption, bool> assumptions;

		private class Assumption
		{
			private readonly Pair<ISchema, ISchema> pair;

			public Assumption(ISchema s, ISchema t)
			{ this.pair = new Pair<ISchema, ISchema>(s, t); }

			public override bool Equals(Object obj)
			{
				if (obj is Assumption) {
					Assumption a = (Assumption) obj;
					bool res = pair.First.StructureEquals(a.pair.First) && pair.Second.StructureEquals(a.pair.Second);
					//if (!res && pair.First.Hash() == a.pair.First.Hash())
					 //Console.WriteLine("===> comparing {0} and {1} = {2}", pair, a.pair, res);
					return res;
				} else
					return false;
			}

			public override int GetHashCode()
			{ return pair.First.Hash() ^ pair.Second.Hash(); }
			//{ return pair.GetHashCode(); }
		}

		public SubtypingChecker(EmptynessChecker ec, DeterminednessChecker dc)
		{
			this.ec = ec;
			this.dc = dc;
			this.handlesOf = new HandlesOfSchemaVisitor(ec);
			this.assumptions = new Dictionary<Assumption, bool>();
		}

		public bool IsSubschema(ISchema s, ISchema t)
		{
 //Console.WriteLine("IN CHECK SUBSCHEMA: {0} <: {1}", s, t);
			Assumption pair = new Assumption(s, t);
			bool res;
			if (assumptions.TryGetValue(pair, out res)) {
 //Console.WriteLine("OUT CHECK SUBSCHEMA: {0} <: {1} = {2} (in assumptions)", s, t, res);
				return res;
			}

			assumptions[pair] = true;
			res = assumptions[pair] = SubSchemaAux(s, t);
 //Console.WriteLine("OUT CHECK SUBSCHEMA: {0} <: {1} = {2}", s, t, res);
			return res;
		}

		private IList<IHandle> GetHandles(ISchema s)
		{ return handlesOf.GetHandles(s); }

		private bool SubSchemaPowerSet(ISchema s1, ISchema s2, IList<ISchema> u1, IList<ISchema> u2, int n, ISchema t1, ISchema t2)
		{
			if (n == 0)
				return IsSubschema(s1, t1) || IsSubschema(s2, t2);
			else
				return SubSchemaPowerSet(s1, s2, u1, u2, n - 1, AddSchema(t1, u1[n - 1]), t2)
					&& SubSchemaPowerSet(s1, s2, u1, u2, n - 1, t1, AddSchema(t2, u2[n - 1]));
		}

		private static bool CapabilityLe(ChannelType.CAPABILITY a, ChannelType.CAPABILITY b)
		{
			return (a == ChannelType.CAPABILITY.INOUT) || (a == b);
		}

		private ISchema AddSchema(ISchema s, ISchema t)
		{ return UnionSchema.Make(s, t); }

		private bool IsSubschema(IHandle hs, IList<IHandle> tHandles)
		{
			if (hs is VoidHandle) {
				foreach (IHandle ht in tHandles)
					if (ht is VoidHandle)
						return true;
				return false;	
			} else if (hs is Handle<BasicSchema>) {
				Handle<BasicSchema> bs = (Handle<BasicSchema>) hs;
				ISchema u = Schema.Empty;
				foreach (IHandle ht in tHandles)
					if (ht is Handle<BasicSchema>) {
						Handle<BasicSchema> bt = (Handle<BasicSchema>) ht;
						if (bs.Head.Type.Kind == bt.Head.Type.Kind)
							u = AddSchema(u, bt.Tail);
					}		
				return IsSubschema(bs.Tail, u);
			} else if (hs is Handle<BasicSchemaLiteralSchema>) {
				Handle<BasicSchemaLiteralSchema> bs = (Handle<BasicSchemaLiteralSchema>) hs;
				ISchema u = Schema.Empty;
				foreach (IHandle ht in tHandles)
					if (ht is Handle<BasicSchema>) {
						Handle<BasicSchema> bt = (Handle<BasicSchema>) ht;
						if (bs.Head.Value.TypeOf().Kind == bt.Head.Type.Kind)
							u = AddSchema(u, bt.Tail);
					} else if (ht is Handle<BasicSchemaLiteralSchema>) {
						Handle<BasicSchemaLiteralSchema> bt = (Handle<BasicSchemaLiteralSchema>) ht;
						if (bs.Head.Value.Equals(bt.Head.Value))
							u = AddSchema(u, bt.Tail);
					}
				return IsSubschema(bs.Tail, u);	
			} else if (hs is Handle<ChannelSchema>) {
				Handle<ChannelSchema> cs = (Handle<ChannelSchema>) hs;
				ISchema u = Schema.Empty;
				foreach (IHandle ht in tHandles)
					if (ht is Handle<ChannelSchema>) {
						Handle<ChannelSchema> ct = (Handle<ChannelSchema>) ht;
						if (CapabilityLe(cs.Head.Capability, ct.Head.Capability)
								// to understand the following conditions just read them
								// as implications:
								// (ct.Head.Capability <= IN) => S <: T (covariance)
								// (ct.Head.Capability <= OUT) => T <: S (contravariance)
							&& (!CapabilityLe(ct.Head.Capability, ChannelType.CAPABILITY.IN) || IsSubschema(cs.Head.Content, ct.Head.Content))
							&& (!CapabilityLe(ct.Head.Capability, ChannelType.CAPABILITY.OUT) || IsSubschema(ct.Head.Content, cs.Head.Content)))
								u = AddSchema(u, ct.Tail);
					}			
				return IsSubschema(cs.Tail, u);
			} else if (hs is Handle<LabelledSchema>) {
				Handle<LabelledSchema> ls = (Handle<LabelledSchema>) hs;
				Queue<ILabelSet> labelsQueue = new Queue<ILabelSet>();
				labelsQueue.Enqueue(ls.Head.Labels);
				while (labelsQueue.Count > 0) {
					ILabelSet labels = labelsQueue.Dequeue();
					IList<ISchema> u1 = new List<ISchema>();
					IList<ISchema> u2 = new List<ISchema>();
					foreach (IHandle ht in tHandles)
						if (ht is Handle<LabelledSchema>) {
							Handle<LabelledSchema> lt = (Handle<LabelledSchema>) ht;
							if (labels.Intersects(lt.Head.Labels) && IsSubschema(ls.Head.Attributes, lt.Head.Attributes)) {
								if (!lt.Head.Labels.Includes(labels)) {
									labelsQueue.Enqueue(LabelSet.Difference(labels, lt.Head.Labels));
									labels = LabelSet.Intersection(labels, lt.Head.Labels);
								}
								u1.Add(lt.Head.Content);
								u2.Add(lt.Tail);
							}
						}	
					if (!SubSchemaPowerSet(ls.Head.Content, ls.Tail, u1, u2, u1.Count, Schema.Empty, Schema.Empty))
						return false;
				}

				return true;
			} 
#if false	//Old code version without subschema relation for record
				else if (hs is Handle<RecordSchema>) {
				Handle<RecordSchema> ss = (Handle<RecordSchema>) hs;
				ISchema u = Schema.Empty;
				foreach (IHandle ht in tHandles)
					if (ht is Handle<RecordSchema>) {
						Handle<RecordSchema> st = (Handle<RecordSchema>) ht;
						bool widthRes = true;
						foreach (KeyValuePair<string, ISchema> field in ss.Head.Fields)
							if (!st.Head.Fields.ContainsKey(field.Key)
								|| !IsSubschema(field.Value, st.Head.Fields[field.Key])) {
								widthRes = false;
								break;
							}
						if (widthRes)
							u = AddSchema(u, st.Tail);
					}
				return IsSubschema(ss.Tail, u);
			} 
#else
				else if (hs is RecordHandle) {
					RecordHandle rs = (RecordHandle) hs;
					ISchema u = Schema.Empty;
					foreach(IHandle ht in tHandles)
						if (ht is RecordHandle) {
							RecordHandle rt = (RecordHandle) ht;
							bool result = true;
							foreach(KeyValuePair<string, IHandle> pair in rt.Fields) {
								if (rs.Fields.ContainsKey(pair.Key)) {
									if (!IsSubschema(rs.Fields[pair.Key].AsSchema(), pair.Value.AsSchema()))
										result = false;
								}
								else
									result = false;
							}
							if (result)
								u = AddSchema(u, rt.Tail);
						}	
					return IsSubschema(rs.Tail, u);						
				}
#endif
				else {
				Debug.Assert(false); // IMPOSSIBLE
				return false;
			}
		}

		private bool SubSchemaAux(ISchema s, ISchema t)
		{
			IList<IHandle> sHandles = GetHandles(s);
			IList<IHandle> tHandles = GetHandles(t);
#if false
			Console.WriteLine(">>>>>>>>>");
			Console.WriteLine("HANDLES OF {0}:", s);
			foreach (IHandle h in sHandles)
				Console.WriteLine("  {0}", h.AsSchema());
			Console.WriteLine("HANDLES OF {0}:", t);
			foreach (IHandle h in tHandles)
				Console.WriteLine("  {0}", h.AsSchema());
			Console.WriteLine("<<<<<<<<<");
#endif
			if (sHandles.Count == 0)
				return true;
			else if (tHandles.Count == 0)
				return false;
			else {
				foreach (IHandle hs in sHandles) {
					if (!IsSubschema(hs, tHandles))
						return false;
				}
			}

			return true;
		}
	}
}
