// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;
using System.Xml;
using System.Xml.Schema;
using PiDuce.Common;

namespace PiDuce.Web
{
    public class SchemaEntry : AbstractEntry
    {
        public SchemaEntry(string name, ISchema schema)
            : base(name, schema)
        { }
    }

    public class IllegalConversionException : Exception
    {
        public IllegalConversionException(string message)
            : base(message)
        { }
    }

    public class XsdImportOutput
    {
        private ISet<string> warnings = new ArraySet<string>();
        private ISet<string> errors = new ArraySet<string>();
        private ISet<string> validationMsg = new ArraySet<string>();

        public void AddWarning(int line, int linePosition, string warning)
        { warnings.Add("line " + line + "," + linePosition + ": " + warning); }

        public void AddWarning(string warning)
        { warnings.Add(warning); }

        public void AddError(int line, int linePosition, string error)
        { errors.Add("line " + line + "," + linePosition + ": " + error); }

        public void AddError(string error)
        { errors.Add(error); }

        public void AddValidationError(string message)
        { validationMsg.Add(message); }

        public ISet<string> Warnings { get { return warnings; } }

        public ISet<string> Errors { get { return errors; } }

        public override String ToString()
        {
            System.Text.StringBuilder builder = new System.Text.StringBuilder();
            foreach (string warning in warnings)
            {
                builder.Append("Warning: " + warning);
                builder.Append(Environment.NewLine);
            }
            foreach (string error in errors)
            {
                builder.Append("Error: " + error);
                builder.Append(Environment.NewLine);
            }
            foreach (string message in validationMsg)
            {
                builder.Append("Validation Error: " + message);
                builder.Append(Environment.NewLine);
            }
            return builder.ToString();
        }
    }

    public class XsdToSchema : IResolver
    {
        public IEntry GetEntry(string name)
        {
            //HACK: here we find unqualified names!
            foreach (XmlQualifiedName qName in this.SchemaDefs.Keys)
            {
                if (qName.Name == name)
                    return this.SchemaDefs[qName].Entry;
            }
            Debug.Assert(false);
            return new SchemaEntry(name, this.SchemaDefs[new XmlQualifiedName(name)]);
        }
        private IDictionary<XmlQualifiedName, ConstantSchema> schemaDefs;
        private IDictionary<XmlQualifiedName, IDictionary<string, ISchema>> attributesSchemaDefs;
        private IDictionary<XmlQualifiedName, ConstantSchema> elementDefs;
        private IDictionary<XmlQualifiedName, ConstantSchema> groupDefs;
        private IDictionary<XmlQualifiedName, ConstantSchema> attributeDefs;

        private IDictionary<XmlQualifiedName, XmlAttribute[]> unhandledSchemaAttributes;
        private IDictionary<XmlQualifiedName, XmlAttribute[]> unhandledElementAttributes;
        private XsdImportOutput output;
        private XmlReader xsd;

        public XsdImportOutput Output { get { return output; } }
        public IDictionary<XmlQualifiedName, ConstantSchema> SchemaDefs { get { return schemaDefs; } }
        public IDictionary<XmlQualifiedName, ConstantSchema> ElementDefs { get { return elementDefs; } }
        public IDictionary<XmlQualifiedName, ConstantSchema> GroupDefs { get { return groupDefs; } }
        public IDictionary<XmlQualifiedName, ConstantSchema> AttrbiuteDefs { get { return attributeDefs; } }
        public IDictionary<XmlQualifiedName, XmlAttribute[]> UnhandledSchemaAttributes { get { return unhandledSchemaAttributes; } }
        public IDictionary<XmlQualifiedName, XmlAttribute[]> UnhandledElementAttributes { get { return unhandledElementAttributes; } }

        public XsdToSchema(XmlReader reader)
        {
            schemaDefs = new Dictionary<XmlQualifiedName, ConstantSchema>();
            attributesSchemaDefs = new Dictionary<XmlQualifiedName, IDictionary<string, ISchema>>();
            elementDefs = new Dictionary<XmlQualifiedName, ConstantSchema>();
            groupDefs = new Dictionary<XmlQualifiedName, ConstantSchema>();
            attributeDefs = new Dictionary<XmlQualifiedName, ConstantSchema>();

            unhandledSchemaAttributes = new Dictionary<XmlQualifiedName, XmlAttribute[]>();
            unhandledElementAttributes = new Dictionary<XmlQualifiedName, XmlAttribute[]>();
            output = new XsdImportOutput();
            xsd = reader;
        }

        public void Compile()
        {
            //This method is divided in three steps: 
            // 1) XmlSchema compilation, 
            // 2) constant schema construction for handling references, 
            // 3) parsing of the compiled schema for building internal schemas (ISchema classes) 
            // 4) constant name resolution

            //step 1
            XmlSchemaSet schemaSet = new XmlSchemaSet();
            ValidationEventHandler vHandler = new ValidationEventHandler(
              delegate(object sender, ValidationEventArgs args)
              { output.AddValidationError(args.Message); });

            schemaSet.ValidationEventHandler += vHandler;
            XmlSchema currentSchema = XmlSchema.Read(xsd, vHandler);
            schemaSet.Add(currentSchema);
            XmlSchemaCompilationSettings settings = new XmlSchemaCompilationSettings();
            settings.EnableUpaCheck = false;
            schemaSet.CompilationSettings = settings;
            schemaSet.XmlResolver = new XmlUrlResolver();
            schemaSet.Compile();

            //step 2
            foreach (XmlSchema schema in schemaSet.Schemas())
                BuildSchemaTable(schema);
            //step 3
            foreach (XmlSchema schema in schemaSet.Schemas())
                Compile(schema);
            //step 4
            ResolveNameSchemaTable();
        }

        private void BuildSchemaTable(XmlSchema schema)
        {
            //stores names contained in schema (it iterates on the inner schemas)
            foreach (XmlQualifiedName name in schema.SchemaTypes.Names)
                schemaDefs[name] = new ConstantSchema(name.Name, null);
            foreach (XmlQualifiedName name in schema.Groups.Names)
                groupDefs[name] = new ConstantSchema(name.Name, null);
            foreach (XmlQualifiedName name in schema.Elements.Names)
                elementDefs[name] = new ConstantSchema(name.Name, null);
            foreach (XmlQualifiedName name in schema.Attributes.Names)
                attributeDefs[name] = new ConstantSchema(name.Name, null);
        }

        private void Compile(XmlSchema schema)
        {
            foreach (XmlQualifiedName name in schema.SchemaTypes.Names)
            {
                ISchema schemaTypeSchema = null;
                IDictionary<string, ISchema> attributes = new Dictionary<string, ISchema>();
                //attributes management
                if (schema.SchemaTypes[name] is XmlSchemaComplexType)
                {
                    XmlSchemaComplexType c = (XmlSchemaComplexType)schema.SchemaTypes[name];
                    attributes = StoreUpAttributes(attributes, c.Attributes, schema.SchemaTypes, schema.Groups, schema.Elements);
                    if (c.ContentModel != null)
                    {
                        if (c.ContentModel.Content is XmlSchemaSimpleContentExtension)
                            attributes = StoreUpAttributes(attributes, ((XmlSchemaSimpleContentExtension)c.ContentModel.Content).Attributes, schema.SchemaTypes, schema.Groups, schema.Elements);
                        if (c.ContentModel.Content is XmlSchemaComplexContentExtension)
                            attributes = StoreUpAttributes(attributes, ((XmlSchemaComplexContentExtension)c.ContentModel.Content).Attributes, schema.SchemaTypes, schema.Groups, schema.Elements);
                    }
                    attributesSchemaDefs[name] = attributes;
                    schemaTypeSchema = VisitXmlSchemaComplexType((XmlSchemaComplexType)schema.SchemaTypes[name], schema.SchemaTypes, schema.Groups, schema.Elements, out attributes);
                }
                else if (schema.SchemaTypes[name] is XmlSchemaSimpleType)
                    schemaTypeSchema = VisitXmlSchemaSimpleType((XmlSchemaSimpleType)schema.SchemaTypes[name], schema.SchemaTypes, schema.Groups, schema.Elements);
                else Debug.Assert(false, "Unexpected class type " + schema.SchemaTypes[name].GetType().ToString());
                SchemaEntry entry = new SchemaEntry(name.Name, schemaTypeSchema);
                schemaDefs[name].Entry = entry;
                unhandledSchemaAttributes[name] = ((XmlSchemaType)schema.SchemaTypes[name]).UnhandledAttributes;
            }

            foreach (XmlQualifiedName name in schema.Attributes.Names)
            {
                ISchema attributeSchema = VisitXmlSchemaAttribute((XmlSchemaAttribute)schema.Attributes[name], schema.SchemaTypes, schema.Groups, schema.Elements);
                SchemaEntry entry = new SchemaEntry(name.Name, attributeSchema);
                attributeDefs[name].Entry = entry;
                unhandledElementAttributes[name] = ((XmlSchemaAttribute)schema.Attributes[name]).UnhandledAttributes;
            }

            foreach (XmlQualifiedName name in schema.Groups.Names)
            {
                ISchema groupSchema = Visit(schema.Groups[name], schema.SchemaTypes, schema.Groups, schema.Elements);
                SchemaEntry entry = new SchemaEntry(name.Name, groupSchema);
                groupDefs[name].Entry = entry;
            }
            foreach (XmlQualifiedName name in schema.Elements.Names)
            {
                ISchema elementSchema = Visit(schema.Elements[name], schema.SchemaTypes, schema.Groups, schema.Elements);
                SchemaEntry entry = new SchemaEntry(name.Name, elementSchema);
                elementDefs[name].Entry = entry;
                unhandledElementAttributes[name] = ((XmlSchemaElement)schema.Elements[name]).UnhandledAttributes;
            }
        }

        private void ResolveNameSchemaTable()
        {
            //Resolves names
            ResolveNamesSchemaVisitor v = new ResolveNamesSchemaVisitor(this);
            foreach (ISchema elementSchema in elementDefs.Values)
                elementSchema.Accept(v);
            foreach (ISchema constantSchema in schemaDefs.Values)
                constantSchema.Accept(v);
            foreach (ISchema constantSchema in groupDefs.Values)
                constantSchema.Accept(v);
            foreach (ISchema constantSchema in attributeDefs.Values)
                constantSchema.Accept(v);
        }

        private ISchema GetBasicSchemaFromName(XmlQualifiedName qName)
        {
            if (!qName.Namespace.Equals(WebNamespace.XmlSchemaNamespace))
                return null;
            else if (qName.Name.Equals("int") || qName.Name.Equals("integer"))
                return new BasicSchema(new IntType());
            else if (qName.Name.Equals("float"))
                return new BasicSchema(new FloatType());
            else if (qName.Name.Equals("boolean"))
                return new BasicSchema(new BoolType());
            else if (qName.Name.Equals("positiveInteger") || qName.Name.Equals("negativeInteger") ||
              qName.Name.Equals("nonPositiveInteger") || qName.Name.Equals("nonNegativeInteger") ||
              qName.Name.Equals("unsignedInt") || qName.Name.Equals("short") || qName.Name.Equals("unsignedShort") ||
              qName.Name.Equals("long") || qName.Name.Equals("unsignedLong"))
            {
                output.AddWarning(qName.Name + "\" imported as integer");
                return new BasicSchema(new IntType());
            }
            else if (qName.Name.Equals("string"))
                return new BasicSchema(new StringType());
            else
            {
                output.AddWarning(qName.Name + "\" imported as string");
                return new BasicSchema(new StringType());
            }
        }

        public ISchema GetSchemaFromName(XmlQualifiedName qName)
        {
            ISchema schema = GetBasicSchemaFromName(qName);
            if (schema != null)
                return schema;
            else if (elementDefs.ContainsKey(qName))
                return elementDefs[qName];
            else if (schemaDefs.ContainsKey(qName))
                return schemaDefs[qName];
            else if (groupDefs.ContainsKey(qName))
                return groupDefs[qName];
            else Debug.Assert(false, qName + " should be defined");
            return null;
        }

        private ISchema BuildRepetition(ISchema t, string minOccurs, string maxOccurs)
        {
            if (minOccurs == null) minOccurs = "1";
            if (maxOccurs == null) maxOccurs = "1";
            uint min = UInt32.Parse(minOccurs);
            if (maxOccurs.Equals("unbounded"))
            {
                if (min == 0) return new StarSchema(t);
                else if (min == 1) return new PlusSchema(t);
                else
                    return new SequenceSchema(new RepetitionSchema(t, min, min), new StarSchema(t));
            }
            else if (min == 1 && UInt32.Parse(maxOccurs) == 1)
                return t;
            else
                return new RepetitionSchema(t, min, UInt32.Parse(maxOccurs));
        }

        private ISchema Visit(XmlSchemaObject t, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            //Console.WriteLine("{0}: Line {1} Col {2}", t, t.LineNumber, t.LinePosition);
            ISchema result = null;
            if (t is XmlSchemaGroupRef)
                result = VisitXmlSchemaGroupRef((XmlSchemaGroupRef)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaGroup)
                result = VisitXmlSchemaGroup((XmlSchemaGroup)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaAny)
                result = VisitXmlSchemaAny((XmlSchemaAny)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaChoice)
                result = VisitXmlSchemaChoice((XmlSchemaChoice)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSequence)
                result = VisitXmlSchemaSequence((XmlSchemaSequence)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaAll)
                result = VisitXmlSchemaAll((XmlSchemaAll)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSimpleType)
                result = VisitXmlSchemaSimpleType((XmlSchemaSimpleType)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaComplexType)
                result = VisitXmlSchemaComplexType((XmlSchemaComplexType)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSimpleTypeRestriction)
                result = VisitXmlSchemaSimpleTypeRestriction((XmlSchemaSimpleTypeRestriction)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSimpleTypeList)
                result = VisitXmlSchemaSimpleTypeList((XmlSchemaSimpleTypeList)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSimpleTypeUnion)
                result = VisitXmlSchemaSimpleTypeUnion((XmlSchemaSimpleTypeUnion)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaComplexContent)
                result = VisitXmlSchemaComplexContent((XmlSchemaComplexContent)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaComplexContentExtension)
                result = VisitXmlSchemaComplexContentExtension((XmlSchemaComplexContentExtension)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaComplexContentRestriction)
                result = VisitXmlSchemaComplexContentRestriction((XmlSchemaComplexContentRestriction)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSimpleContent)
                result = VisitXmlSchemaSimpleContent((XmlSchemaSimpleContent)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSimpleContentExtension)
                result = VisitXmlSchemaSimpleContentExtension((XmlSchemaSimpleContentExtension)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaElement)
                result = VisitXmlSchemaElement((XmlSchemaElement)t, schemaDefs, groupDefs, elementDefs);
            else if (t is XmlSchemaSimpleContentRestriction)
                result = VisitXmlSchemaSimpleContentRestriction((XmlSchemaSimpleContentRestriction)t, schemaDefs, groupDefs, elementDefs);
            else
            {
                output.AddError(t.LineNumber, t.LinePosition, "\"" + t.GetType() + "\" is not supported");
                throw new IllegalConversionException(output.ToString());
            }
            return result;
        }

        private ISchema VisitXmlSchemaAny(XmlSchemaAny s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        { return Schema.Any; }

        private ISchema VisitXmlSchemaChoice(XmlSchemaChoice s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            ISchema content = null;
            if (s.Items.Count == 0)
                content = Schema.Empty;
            foreach (XmlSchemaObject t in s.Items)
            {
                if (content == null) content = Visit(t, schemaDefs, groupDefs, elementDefs);
                else content = UnionSchema.Make(content, Visit(t, schemaDefs, groupDefs, elementDefs));
            }
            return BuildRepetition(content, s.MinOccursString, s.MaxOccursString);
        }

        private ISchema VisitXmlSchemaSequence(XmlSchemaSequence s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            ISchema content = null;
            if (s.Items.Count == 0)
                content = Schema.Empty;
            foreach (XmlSchemaObject t in s.Items)
            {
                if (content == null) content = Visit(t, schemaDefs, groupDefs, elementDefs);
                else content = new SequenceSchema(content, Visit(t, schemaDefs, groupDefs, elementDefs));
            }
            return BuildRepetition(content, s.MinOccursString, s.MaxOccursString);
        }

        private ISchema VisitXmlSchemaAll(XmlSchemaAll s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            output.AddWarning(s.LineNumber, s.LinePosition, "\"all\" imported as an ordered sequence");
            ISchema content = null;
            if (s.Items.Count == 0)
                content = Schema.Empty;
            foreach (XmlSchemaObject t in s.Items)
            {
                if (content == null) content = Visit(t, schemaDefs, groupDefs, elementDefs);
                else content = new AllSchema(content, Visit(t, schemaDefs, groupDefs, elementDefs));
            }
            return BuildRepetition(content, s.MinOccursString, s.MaxOccursString);
        }

        private ISchema VisitXmlSchemaSimpleType(XmlSchemaSimpleType s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            /*
             * <simpleType
             * final = (#all | (list | union | restriction))
             * id = ID
             * name = NCName
             * {any attributes with non-schema namespace . . .}>
             * Content: (annotation?, (restriction | list | union))
             * </simpleType>
             */
            if (s.Final != XmlSchemaDerivationMethod.Empty && s.Final != XmlSchemaDerivationMethod.None)
                output.AddWarning(s.LineNumber, s.LinePosition, "\"final simpleType\" is ignored");
            if (s.IsMixed)
                output.AddWarning(s.LineNumber, s.LinePosition, "\"mixed simpleType\" is ignored");
            return Visit(s.Content, schemaDefs, groupDefs, elementDefs);
        }

        private ISchema VisitXmlSchemaAttribute(XmlSchemaAttribute s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            ISchema schema = null;
            if (!s.RefName.IsEmpty)
                schema = attributeDefs[s.RefName];
            else if (s.SchemaTypeName != null && !s.SchemaTypeName.IsEmpty)
                schema = GetSchemaFromName(s.SchemaTypeName);
            else
                schema = Visit(s.SchemaType, schemaDefs, groupDefs, elementDefs);

            switch (s.Use) //the current use overwrites the one declare in the ref
            {
                case XmlSchemaUse.Optional:
                    return new RepetitionSchema(schema, 0, 1);
                case XmlSchemaUse.Required:
                    return schema;
                case XmlSchemaUse.None:
                    return schema;
                case XmlSchemaUse.Prohibited:
                    throw new ApplicationException("VisitXmlSchemaAttribute cannot be called with a Prohibited use");
            }
            return schema;
        }

        private ISchema VisitXmlSchemaSimpleTypeRestriction(XmlSchemaSimpleTypeRestriction s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            /* <restriction
             * base = QName
             * id = ID
             * {any attributes with non-schema namespace . . .}>
             * Content: (annotation?, (simpleType?, (minExclusive | minInclusive | maxExclusive | maxInclusive | 
             * totalDigits | fractionDigits | length | minLength | maxLength | enumeration | whiteSpace | pattern)*))
             * </restriction>
             */
            ISchema content = null;
            content = GetSchemaFromName(s.BaseTypeName);
            if (content is ConstantSchema)
                content = Visit(schemaDefs[s.BaseTypeName], schemaDefs, groupDefs, elementDefs);
            bool enumeration = false;
            bool length = false;
            foreach (XmlSchemaObject f in s.Facets)
            {
                if (f is XmlSchemaEnumerationFacet)
                    enumeration = true;
                if (f is XmlSchemaMinLengthFacet || f is XmlSchemaMaxLengthFacet ||
                  f is XmlSchemaLengthFacet)
                    length = true;
            }
            string minOccurs = "1";
            string maxOccurs = "1";
            ISchema argumentSchema;
            if (content is StarSchema)
            {
                argumentSchema = ((StarSchema)content).Content;
                minOccurs = "0";
                maxOccurs = "unbounded";
            }
            else if (content is PlusSchema)
            {
                argumentSchema = ((PlusSchema)content).Content;
                minOccurs = "1";
                maxOccurs = "unbounded";
            }
            else if (content is RepetitionSchema)
            {
                argumentSchema = ((RepetitionSchema)content).Content;
                minOccurs = ((RepetitionSchema)content).MinOccurs.ToString();
                maxOccurs = ((RepetitionSchema)content).MaxOccurs.ToString();
            }
            else argumentSchema = content;

            if (enumeration)
            {
                ISchema choice = null;
                foreach (XmlSchemaObject f in s.Facets)
                {
                    if (f is XmlSchemaEnumerationFacet)
                    {
                        if (choice == null)
                            choice = VisitXmlSchemaEnumerationFacet((XmlSchemaEnumerationFacet)f, argumentSchema);
                        else
                            choice = UnionSchema.Make(choice,
                              VisitXmlSchemaEnumerationFacet((XmlSchemaEnumerationFacet)f, argumentSchema));
                    }
                }
                argumentSchema = choice;
            }

            if (length && !(content is StarSchema || content is PlusSchema || content is RepetitionSchema))
                output.AddWarning(s.LineNumber, s.LinePosition, "XmlSchema LengthFacet is not supported for non-list types " + content.GetType());
            else if (length)
            {
                foreach (XmlSchemaObject f in s.Facets)
                {
                    if (f is XmlSchemaPatternFacet || f is XmlSchemaWhiteSpaceFacet) //todo Whitespace
                        output.AddWarning(f.LineNumber, f.LinePosition, f.GetType() + "is not supported");
                }
                foreach (XmlSchemaObject f in s.Facets)
                {
                    if (f is XmlSchemaLengthFacet)
                    {
                        Debug.Assert(maxOccurs.Equals("unbounded") || UInt32.Parse(maxOccurs) >= UInt32.Parse(((XmlSchemaLengthFacet)f).Value));
                        Debug.Assert(UInt32.Parse(minOccurs) <= UInt32.Parse(((XmlSchemaLengthFacet)f).Value));
                        minOccurs = maxOccurs = ((XmlSchemaLengthFacet)f).Value;
                    }
                    else if (f is XmlSchemaMinLengthFacet)
                    {
                        Debug.Assert(UInt32.Parse(minOccurs) <= UInt32.Parse(((XmlSchemaMinLengthFacet)f).Value));
                        minOccurs = ((XmlSchemaMinLengthFacet)f).Value;
                    }
                    else if (f is XmlSchemaMaxLengthFacet)
                    {
                        Debug.Assert(maxOccurs.Equals("unbounded") || UInt32.Parse(maxOccurs) >= UInt32.Parse(((XmlSchemaMaxLengthFacet)f).Value));
                        maxOccurs = ((XmlSchemaMaxLengthFacet)f).Value;
                    }
                }
            }
            return BuildRepetition(argumentSchema, minOccurs, maxOccurs);
        }

        private ISchema VisitXmlSchemaSimpleTypeList(XmlSchemaSimpleTypeList s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            /*
             * <list
             * id = ID
             * itemType = QName
             * {any attributes with non-schema namespace . . .}>
             * Content: (annotation?, (simpleType?))
             * </list>
             * Possible restrictions: length | maxLength | minLength | enumeration | pattern | whiteSpace
             */
            ISchema content = null;
            if (!s.ItemTypeName.IsEmpty)
                content = GetSchemaFromName(s.ItemTypeName);
            else
                content = Visit(s.BaseItemType, schemaDefs, groupDefs, elementDefs);
            string minOccurs = "0";
            string maxOccurs = "unbounded";
            if (s.ItemType != null && s.ItemType.Content is XmlSchemaSimpleTypeRestriction)
            {
                XmlSchemaSimpleTypeRestriction restriction = (XmlSchemaSimpleTypeRestriction)s.ItemType.Content;
                foreach (XmlSchemaObject f in restriction.Facets)
                {
                    if (f is XmlSchemaPatternFacet || f is XmlSchemaWhiteSpaceFacet) //todo Whitespace
                        output.AddWarning(f.LineNumber, f.LinePosition, f.GetType() + "is not supported");
                }
                foreach (XmlSchemaObject f in restriction.Facets)
                {
                    if (f is XmlSchemaLengthFacet)
                        minOccurs = maxOccurs = ((XmlSchemaLengthFacet)f).Value;
                    else if (f is XmlSchemaMinLengthFacet)
                        minOccurs = ((XmlSchemaMinLengthFacet)f).Value;
                    else if (f is XmlSchemaMaxLengthFacet)
                        maxOccurs = ((XmlSchemaMaxLengthFacet)f).Value;
                }
            }
            return BuildRepetition(content, minOccurs, maxOccurs);
        }

        private ISchema VisitXmlSchemaSimpleTypeUnion(XmlSchemaSimpleTypeUnion s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            ISchema content = null;
            if (s.MemberTypes != null)
            {
                foreach (XmlQualifiedName baseTypeName in s.MemberTypes)
                {
                    if (content == null) content = GetSchemaFromName(baseTypeName);
                    else content = UnionSchema.Make(content, GetSchemaFromName(baseTypeName));
                }
            }
            if (s.BaseMemberTypes != null)
            {
                foreach (XmlSchemaSimpleType baseType in s.BaseMemberTypes)
                {
                    if (content == null)
                        content = VisitXmlSchemaSimpleType(baseType, schemaDefs, groupDefs, elementDefs);
                    else if (baseType.QualifiedName.Equals(new XmlQualifiedName()))
                        content = UnionSchema.Make(content, VisitXmlSchemaSimpleType(baseType, schemaDefs, groupDefs, elementDefs));
                }
            }
            return content;
        }

        private ISchema VisitXmlSchemaComplexContent(XmlSchemaComplexContent s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            if (s.IsMixed)
                output.AddWarning(s.LineNumber, s.LinePosition, "\"mixed content model\" is ignored");
            return Visit(s.Content, schemaDefs, groupDefs, elementDefs);
        }

        private ISchema VisitXmlSchemaComplexContentExtension(XmlSchemaComplexContentExtension s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            if (s.Particle != null)
            {
                ISchema tail = Visit(s.Particle, schemaDefs, groupDefs, elementDefs);
                Debug.Assert(!s.BaseTypeName.IsEmpty);
                ISchema head = GetSchemaFromName(s.BaseTypeName);
                return new SequenceSchema(head, tail);
            }
            else
                return (GetSchemaFromName(s.BaseTypeName));
        }

        private ISchema VisitXmlSchemaComplexContentRestriction(XmlSchemaComplexContentRestriction s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            //complex content restriction allows to add attributes
            XmlQualifiedName arrayQName = new XmlQualifiedName("Array", WebNamespace.SoapEncNamespace);
            if (s.BaseTypeName.Equals(arrayQName))
            {
                //<attribute ref="soapenc:arrayType" wsdl:arrayType="xsd:float[]"/> 
                XmlSchemaAttribute attr = (XmlSchemaAttribute)s.Attributes[0];
                string name = String.Empty;
                string namespaceUri = String.Empty;
                string[] size = null;
                foreach (XmlAttribute a in attr.UnhandledAttributes)
                {
                    if (a.NamespaceURI == WebNamespace.WsdlNamespace && a.LocalName == "arrayType")
                    {
                        string schemaName = a.Value.Substring(0, a.Value.IndexOf('['));
                        int pos = schemaName.LastIndexOf(':');
                        if (pos > 0)
                        {
                            namespaceUri = schemaName.Substring(0, pos);
                            //this is for MONO that lookups the namespace automatically (.Net does not)
                            try
                            { new Uri(namespaceUri, UriKind.Absolute); }
                            catch (Exception)
                            { namespaceUri = xsd.LookupNamespace(namespaceUri); }
                            name = schemaName.Substring(pos + 1);
                        }
                        else name = schemaName;
                        string boundaries = a.Value.Substring(a.Value.IndexOf('['));
                        if (boundaries.Contains(","))
                            this.output.AddError(s.LineNumber, s.LinePosition, "Multi-dimensional arrays are not supported (only jagged array are supported");
                        boundaries = boundaries.Replace(",", "[]");
                        Regex e = new Regex(@"^\[{[0-9] }*\]$");
                        size = e.Split(boundaries);
                    }
                }
                XmlQualifiedName qName = new XmlQualifiedName(name, namespaceUri);
                ISchema arraySchema = null;
                int j = size.Length - 1;
                do
                {
                    size[j] = size[j].Replace(" ", "").Replace("[", "").Replace("]", "");
                    if (size[j] == "")
                    {
                        if (arraySchema == null)
                            arraySchema = new ArrayStarSchema(qName, new LabelledSchema(LabelSet.Singleton("Item"), GetSchemaFromName(qName)));
                        else new ArrayStarSchema(qName, new LabelledSchema(LabelSet.Singleton("Item"), arraySchema));
                    }
                    else
                    {
                        int occurs = int.Parse(size[j]);
                        if (arraySchema == null)
                            arraySchema = new ArrayRepetitionSchema(qName, new LabelledSchema(LabelSet.Singleton("Item"), GetSchemaFromName(qName)), (uint)occurs);
                        else new ArrayRepetitionSchema(qName, new LabelledSchema(LabelSet.Singleton("Item"), arraySchema), (uint)occurs);
                    }
                    j--;
                } while (j > 0);
                return arraySchema;
            }
            else if (s.Particle == null)
            {
                output.AddError(s.LineNumber, s.LinePosition, "\"complex content restriction\" is ignored");
                return new VoidSchema();
            }
            else
                return Visit(s.Particle, schemaDefs, groupDefs, elementDefs);
        }

        private ISchema VisitXmlSchemaGroupRef(XmlSchemaGroupRef s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            return GetSchemaFromName(s.RefName);
        }

        private ISchema VisitXmlSchemaGroup(XmlSchemaGroup s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            return Visit(s.Particle, schemaDefs, groupDefs, elementDefs);
        }

        //read a complex type without considering attributes
        private ISchema VisitXmlSchemaComplexType(XmlSchemaComplexType s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            //attributes are read in VisitXmlSchemaElement and in the top level parsing of schema definitions
            IDictionary<string, ISchema> foo;
            return VisitXmlSchemaComplexType(s, schemaDefs, groupDefs, elementDefs, out foo);
        }

        //read a complex type considering attributes
        private ISchema VisitXmlSchemaComplexType(XmlSchemaComplexType s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs, out IDictionary<string, ISchema> attributes)
        {
            attributes = new Dictionary<string, ISchema>();
            attributes = StoreUpAttributes(attributes, s.Attributes, schemaDefs, groupDefs, elementDefs);
            if (s.ContentModel != null)
            {
                if (s.ContentModel.Content is XmlSchemaSimpleContentExtension)
                    attributes = StoreUpAttributes(attributes, ((XmlSchemaSimpleContentExtension)s.ContentModel.Content).Attributes, schemaDefs, groupDefs, elementDefs);
                if (s.ContentModel.Content is XmlSchemaComplexContentExtension)
                    attributes = StoreUpAttributes(attributes, ((XmlSchemaComplexContentExtension)s.ContentModel.Content).Attributes, schemaDefs, groupDefs, elementDefs);
            }
            if (s.IsMixed)
                output.AddWarning(s.LineNumber, s.LinePosition, "\"mixed content model\" is ignored");
            if (s.IsAbstract)
                output.AddWarning(s.LineNumber, s.LinePosition, "\"abstract complexType\" is ignored");
            if (s.Final != XmlSchemaDerivationMethod.Empty && s.Final != XmlSchemaDerivationMethod.None)
                output.AddWarning(s.LineNumber, s.LinePosition, "\"final complexType\" is ignored");
            if (s.ContentModel != null)
                return Visit(s.ContentModel, schemaDefs, groupDefs, elementDefs);
            else if (s.Particle == null)
                return new VoidSchema();
            else
                return Visit(s.Particle, schemaDefs, groupDefs, elementDefs);
        }

        private ISchema VisitXmlSchemaSimpleContent(XmlSchemaSimpleContent s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        { return Visit(s.Content, schemaDefs, groupDefs, elementDefs); }

        private ISchema VisitXmlSchemaSimpleContentExtension(XmlSchemaSimpleContentExtension s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            output.AddError(s.LineNumber, s.LinePosition, "\"simpleType extension\" is not supported");
            return GetSchemaFromName(s.BaseTypeName);
        }

        private bool IsWithoutElement(XmlSchemaElement s, out ISet<string> labels)
        {
            labels = new ArraySet<string>();
            if (s.UnhandledAttributes == null) return false;
            bool isWithout = false;
            foreach (XmlAttribute a in s.UnhandledAttributes)
                if (a.NamespaceURI.Equals(WebNamespace.PiDuceNamespace) &&
                  a.LocalName.Equals("schema") && a.Value.Equals("without"))
                {
                    isWithout = true;
                    if (s.Annotation != null)
                        foreach (XmlSchemaObject annotation in s.Annotation.Items)
                            if (annotation is XmlSchemaAppInfo)
                                foreach (XmlNode without in ((XmlSchemaAppInfo)annotation).Markup)
                                    labels.Add(without.Attributes["label"].Value);
                }
            return isWithout;
        }

        private bool IsPiDuceElement(XmlSchemaElement s, out ISchema schema)
        {
            if (s.UnhandledAttributes == null)
            {
                schema = null;
                return false;
            }
            foreach (XmlAttribute a in s.UnhandledAttributes)
            {
                if (a.NamespaceURI.Equals(WebNamespace.PiDuceNamespace) && a.LocalName.Equals("schema") &&
                  a.Value.Equals("PiDuceChannel"))
                {
                    ChannelType.CAPABILITY capability = ChannelType.CAPABILITY.OUT;
                    if (a.NamespaceURI.Equals(WebNamespace.PiDuceNamespace) && a.LocalName.Equals("capability"))
                    {
                        if (a.Value.Equals("I"))
                            capability = ChannelType.CAPABILITY.IN;
                        else if (a.Value.Equals("O"))
                            capability = ChannelType.CAPABILITY.OUT;
                        else if (a.Value.Equals("IO"))
                            capability = ChannelType.CAPABILITY.INOUT;
                    }
                    foreach (XmlSchemaObject annotation in s.Annotation.Items)
                    {
                        if (annotation is XmlSchemaAppInfo)
                        {
                            XmlNode serviceInterface = ((XmlSchemaAppInfo)annotation).Markup[0];
                            XmlNamespaceManager nsmgr = new XmlNamespaceManager(xsd.NameTable);
                            nsmgr.AddNamespace("piduce", WebNamespace.PiDuceNamespace);
                            nsmgr.AddNamespace("xsd", WebNamespace.XmlSchemaNamespace);
                            XmlReader xsdSchema = new XmlNodeReader(serviceInterface.SelectSingleNode("/piduce:interface/piduce:types/xsd:schema", nsmgr));
                            XsdToSchema tmp = new XsdToSchema(xsdSchema);
                            tmp.Compile();
                            ISchema inputSchema = null;
                            ISchema outputSchema = null;
                            XmlNode inputNode = serviceInterface.SelectSingleNode("/piduce:interface/piduce:input", nsmgr);
                            if (inputNode != null)
                            {
                                string messageName = inputNode.Attributes["element", WebNamespace.PiDuceNamespace].Value;
                                if (messageName.IndexOf(':') != -1)
                                    messageName = messageName.Substring(messageName.IndexOf(':') + 1);
                                XmlQualifiedName n = new XmlQualifiedName(messageName, WebNamespace.PiDuceDefaultNamespace);
                                bool isWrapper = false;
                                if (tmp.UnhandledElementAttributes[n].Length > 0)
                                {
                                    foreach (XmlAttribute att in tmp.UnhandledElementAttributes[n])
                                    {
                                        if (att.LocalName == "isWrapper" && att.Value == "yes" && att.NamespaceURI == WebNamespace.PiDuceNamespace)
                                            isWrapper = true;
                                    }
                                }
                                inputSchema = tmp.ElementDefs[n];
                                if (isWrapper)
                                    inputSchema = ((LabelledSchema)inputSchema).Content;
                            }
                            XmlNode outputNode = serviceInterface.SelectSingleNode("/piduce:interface/piduce:output", nsmgr);
                            if (outputNode != null)
                            {
                                string messageName = outputNode.Attributes["element", WebNamespace.PiDuceNamespace].Value;
                                if (messageName.IndexOf(':') != -1)
                                    messageName = messageName.Substring(messageName.IndexOf(':') + 1);
                                XmlQualifiedName n = new XmlQualifiedName(messageName, WebNamespace.PiDuceDefaultNamespace);
                                bool isWrapper = false;
                                if (tmp.UnhandledElementAttributes[n].Length > 0)
                                {
                                    foreach (XmlAttribute att in tmp.UnhandledElementAttributes[n])
                                    {
                                        if (att.LocalName == "isWrapper" && att.Value == "yes" && att.NamespaceURI == WebNamespace.PiDuceNamespace)
                                            isWrapper = true;
                                    }
                                }
                                outputSchema = tmp.ElementDefs[n];
                                if (isWrapper)
                                    outputSchema = ((LabelledSchema)outputSchema).Content;
                            }
                            if (outputSchema == null) schema = new ChannelSchema(inputSchema, capability);
                            else schema = new FunctionSchema(inputSchema, outputSchema);
                            return true;
                        }
                    }
                }
                else if (a.NamespaceURI.Equals(WebNamespace.PiDuceNamespace) && a.LocalName.Equals("schema") &&
                  a.Value.Equals("PiDuceService"))
                {
                    foreach (XmlSchemaObject annotation in s.Annotation.Items)
                    {
                        if (annotation is XmlSchemaAppInfo)
                        {
                            string wsdl = ((XmlSchemaAppInfo)annotation).Markup[0].OuterXml;
                            XsdToSchema converter;
                            schema = AbstractWsdlReader.CompileAbstractWsdl(wsdl, out converter);
                            return true;
                        }
                    }
                }
            }
            schema = null;
            return false;
        }

        private ISchema VisitXmlSchemaElement(XmlSchemaElement s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            /*
             * <element
             * abstract = boolean : false
             * block = (#all | List of (extension | restriction | substitution))
             * default = string
             * final = (#all | List of (extension | restriction))
             * fixed = string
             * form = (qualified | unqualified)
             * id = ID
             * maxOccurs = (nonNegativeInteger | unbounded)  : 1
             * minOccurs = nonNegativeInteger : 1
             * name = NCName
             * nillable = boolean : false
             * ref = QName
             * substitutionGroup = QName
             * schema = QName
             * {any attributes with non-schema namespace . . .}>
             * Content: (annotation?, ((simpleType | complexType)?, (unique | key | keyref)*))
             * </element>
             */
            ISchema schema;
            if (IsPiDuceElement(s, out schema))
                return schema;
            else if (!s.RefName.IsEmpty)
                return GetSchemaFromName(s.RefName);
            else
            {
                ISchema content = null;
                IDictionary<string, ISchema> attributes = null;
                string name = s.Name;
                if (s.Block != XmlSchemaDerivationMethod.Empty && s.Block != XmlSchemaDerivationMethod.None)
                    output.AddWarning(s.LineNumber, s.LinePosition, "\"block\" is ignored");
                if (s.IsAbstract)
                    output.AddWarning(s.LineNumber, s.LinePosition, "\"abstract\" is ignored");
                if (s.Form == XmlSchemaForm.Qualified)
                    output.AddWarning(s.LineNumber, s.LinePosition, "\"qualified form\" is not supported");
                if (!s.SchemaTypeName.IsEmpty)
                {
                    content = GetSchemaFromName(s.SchemaTypeName);
                    if (!attributesSchemaDefs.TryGetValue(s.SchemaTypeName, out attributes))
                      attributes = new Dictionary<string, ISchema>();
                }
                else if (!s.SubstitutionGroup.IsEmpty)
                    content = GetSchemaFromName(s.SubstitutionGroup);
                if (content == null)
                {
                    if (s.SchemaType == null)
                    {
                        content = new VoidSchema();
                        attributes = new Dictionary<string, ISchema>();
                    }
                    else if (s.SchemaType is XmlSchemaComplexType)
                        content = VisitXmlSchemaComplexType((XmlSchemaComplexType)s.SchemaType, schemaDefs, groupDefs, elementDefs, out attributes);
                    else if (s.SchemaType is XmlSchemaSimpleType)
                    {
                        content = VisitXmlSchemaSimpleType((XmlSchemaSimpleType)s.SchemaType, schemaDefs, groupDefs, elementDefs);
                        attributes = new Dictionary<string, ISchema>();
                    }
                    else
                        Debug.Assert(false, "Unexpected class type " + s.SchemaType.GetType().ToString());
                }
                Debug.Assert(content != null);
                if (s.FixedValue != null && s.FixedValue.Length != 0)
                {
                    if (content is BasicSchema)
                    {
                        IBasicType type = ((BasicSchema)content).Type;
                        if (type is IntType)
                            content = new BasicSchemaLiteralSchema(new IntLiteral(Int32.Parse(s.FixedValue)));
                        else if (type is StringType)
                            content = new BasicSchemaLiteralSchema(new StringLiteral(s.FixedValue));
                        else
                            output.AddWarning(s.LineNumber, s.LinePosition, "\"fixed\" is supported only for basic types");
                    }
                }
                if (s.DefaultValue != null && s.DefaultValue.Length != 0)
                    output.AddWarning(s.LineNumber, s.LinePosition, "\"default\" is ignored");

                ISet<string> labels;
                if (IsWithoutElement(s, out labels))
                    content = new LabelledSchema(LabelSet.Without(labels), content, new RecordSchema(attributes));
                else
                    content = new LabelledSchema(LabelSet.Singleton(name), content, new RecordSchema(attributes));
                if (s.IsNillable)
                    content = UnionSchema.Make(content, new VoidSchema());
                //Console.WriteLine("</{0}>", s.QualifiedName);
                return schema = BuildRepetition(content, s.MinOccursString, s.MaxOccursString);
            }
        }

        private ISchema VisitXmlSchemaEnumerationFacet(XmlSchemaEnumerationFacet s, ISchema baseSchema)
        {
            if (baseSchema is BasicSchema)
            {
                IBasicType type = ((BasicSchema)baseSchema).Type;
                if (type is IntType)
                    return new BasicSchemaLiteralSchema(new IntLiteral(Int32.Parse(s.Value)));
                else if (type is StringType)
                    return new BasicSchemaLiteralSchema(new StringLiteral(s.Value));
            }

            output.AddWarning(s.LineNumber, s.LinePosition, "\"Enumeration\" on " +
              baseSchema + " is ignored (supported only on basic types)");
            return null;
        }

        private ISchema VisitXmlSchemaSimpleContentRestriction(XmlSchemaSimpleContentRestriction s, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            /* <restriction
             * base = QName
             * id = ID
             * {any attributes with non-schema namespace . . .}>
             * Content: (annotation?, (simpleType?, (minExclusive | minInclusive | maxExclusive | maxInclusive | totalDigits 
             * | fractionDigits | length | minLength | maxLength | enumeration | whiteSpace | pattern)*)?, 
             * ((attribute | attributeGroup)*, anyAttribute?))
             * </restriction>
             */
            ISchema content = null;
            if (s.Facets.Count == 0 && !s.BaseTypeName.IsEmpty)
                content = GetSchemaFromName(s.BaseTypeName);
            else if (s.Facets.Count == 0)
                content = Visit(s.BaseType, schemaDefs, groupDefs, elementDefs);

            bool first = true;
            foreach (XmlSchemaObject f in s.Facets)
            {
                if (f is XmlSchemaEnumerationFacet)
                {
                    if (first)
                    {
                        content = VisitXmlSchemaEnumerationFacet((XmlSchemaEnumerationFacet)f, content);
                        first = false;
                    }
                    else content = UnionSchema.Make(content, VisitXmlSchemaEnumerationFacet((XmlSchemaEnumerationFacet)f, content));
                }
                else
                    output.AddWarning(f.LineNumber, f.LinePosition, "only enumeration is supported");
            }
            return content;
        }

        private IDictionary<string, ISchema> StoreUpAttributes(IDictionary<string, ISchema> map, XmlSchemaObjectCollection attributes, XmlSchemaObjectTable schemaDefs, XmlSchemaObjectTable groupDefs, XmlSchemaObjectTable elementDefs)
        {
            foreach (XmlSchemaObject obj in attributes)
            {
                if (obj is XmlSchemaAttribute)
                {
                    XmlSchemaAttribute attr = (XmlSchemaAttribute)obj;
                    string attrName = null;
                    if (!attr.RefName.IsEmpty) attrName = attr.RefName.Name;
                    else attrName = attr.Name;
                    if (attr.Use != XmlSchemaUse.Prohibited)
                        map.Add(attrName, VisitXmlSchemaAttribute(attr, schemaDefs, groupDefs, elementDefs));
                }
            }
            return map;
        }
    }
}